/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS3

options {
	query-source address 10.53.0.7;
	notify-source 10.53.0.7;
	transfer-source 10.53.0.7;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.7; };
	listen-on-v6 { none; };
	recursion no;
	notify yes;
	dnssec-validation yes;
	minimal-responses yes;
};

zone "." {
	type hint;
	file "../../_common/root.hint";
};

zone "nsec3.example" {
	type secondary;
	primaries { 10.53.0.3; };
	file "nsec3.example.bk";
};

zone "optout.example" {
	type secondary;
	primaries { 10.53.0.3; };
	file "optout.example.bk";
};

zone "nsec3-unknown.example" {
	type secondary;
	primaries { 10.53.0.3; };
	file "nsec3-unknown.example.bk";
};

zone "optout-unknown.example" {
	type secondary;
	primaries { 10.53.0.3; };
	file "optout-unknown.example.bk";
};

zone "multiple.example" {
	type secondary;
	primaries { 10.53.0.3; };
	file "multiple.example.bk";
};

zone "nosoa.secure.example" {
	type primary;
	file "nosoa.secure.example.db";
};

zone "split-rrsig" {
	type primary;
	file "split-rrsig.db.signed";
	allow-update { any; };
};

include "trusted.conf";
