/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

/*
 * Test rpz performance.
 */

options {
	query-source address 10.53.0.5;
	notify-source 10.53.0.5;
	transfer-source 10.53.0.5;
	port @PORT@;
	pid-file "named.pid";
	statistics-file "named.stats";
	session-keyfile "session.key";
	listen-on { 10.53.0.5; };
	listen-on-v6 { none; };
	allow-transfer { any; };
	ixfr-from-differences yes;
	notify-delay 0;
	notify yes;
	minimal-responses no;
	recursion yes;
	dnssec-validation yes;

	# turn rpz on or off
	include "rpz-switch";

	include "../dnsrps.conf";
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};
controls {
	inet 10.53.0.5 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

include "../trusted.conf";
zone "."		{type hint; file "hints"; };

zone "tld5."		{type primary; file "tld5.db"; };
zone "example.tld5."	{type primary; file "example.db"; };

zone "bl0."		{type primary; file "bl.db"; };
zone "bl1."		{type primary; file "bl.db"; };
zone "bl2."		{type primary; file "bl.db"; };
zone "bl3."		{type primary; file "bl.db"; };
zone "bl4."		{type primary; file "bl.db"; };
zone "bl5."		{type primary; file "bl.db"; };
zone "bl6."		{type primary; file "bl.db"; };
zone "bl7."		{type primary; file "bl.db"; };
zone "bl8."		{type primary; file "bl.db"; };
zone "bl9."		{type primary; file "bl.db"; };
zone "bl10."		{type primary; file "bl.db"; };
zone "bl11."		{type primary; file "bl.db"; };
zone "bl12."		{type primary; file "bl.db"; };
zone "bl13."		{type primary; file "bl.db"; };
zone "bl14."		{type primary; file "bl.db"; };
zone "bl15."		{type primary; file "bl.db"; };
zone "bl16."		{type primary; file "bl.db"; };
zone "bl17."		{type primary; file "bl.db"; };
zone "bl18."		{type primary; file "bl.db"; };
zone "bl19."		{type primary; file "bl.db"; };

zone "policy1" {
	type primary;
	file "empty.db";
	also-notify { 10.53.0.6; };
	allow-update { any; };
	allow-transfer { any; };
};

zone "policy2" {
	type primary;
	file "policy2.db";
	allow-update { any; };
	allow-transfer { any; };
};

include "expire.conf";
