/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

dnssec-policy "autosign" {

	signatures-refresh P1W;
	signatures-validity P2W;
	signatures-validity-dnskey P2W;

	dnskey-ttl 300;

	keys {
		ksk key-directory lifetime P2Y algorithm @DEFAULT_ALGORITHM@;
		zsk key-directory lifetime P1Y algorithm @DEFAULT_ALGORITHM@;
	};
};

dnssec-policy "manual" {

	signatures-refresh P1W;
	signatures-validity P2W;
	signatures-validity-dnskey P2W;

	dnskey-ttl 300;

	keys {
		ksk key-directory lifetime P2Y algorithm @DEFAULT_ALGORITHM@;
		zsk key-directory lifetime P2M algorithm @DEFAULT_ALGORITHM@;
	};

	manual-mode yes;
};
