/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

{% if FEATURE_FIPS_DH == "1" %}
include "named-tls.conf";
{% endif %}

options {
	query-source address 10.53.0.3;
	notify-source 10.53.0.3;
	transfer-source 10.53.0.3;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.3; };
	listen-on-v6 { fd92:7065:b8e:ffff::3; };
	recursion yes;
	notify yes;
	dnssec-validation no;

{% if FEATURE_FIPS_DH == "1" %}
	tls-port @TLSPORT@;
{% endif %}

};

zone "." {
	type hint;
	file "../../_common/root.hint";
};

zone "example" {
	type secondary;
	primaries { 10.53.0.2; };
	file "example.bk";
	# non-responsive notify recipient (no reply, no ICMP errors)
	also-notify { 10.53.10.53; };
};

zone "notify-source-port-test" {
	type primary;
	file "notify-source-port-test.db";
	notify-source 10.53.0.3 port @EXTRAPORT2@;
};
