/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

acl accept { 10.53.0.2; };

options {
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { none; };
	recursion no;
	dnssec-validation no;
};

include "controls.conf";

view "internal" {

	allow-query {! accept; };

	zone "." {
		type hint;
		file "../../_common/root.hint";
	};

	zone "normal.example" {
		type primary;
		file "generic.db";
	};
};
