/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

/*
 * The zone in view "example1" inherits the key directory value from "options",
 * but in view "example2" sets the key directory to the same value. This should
 * be detected as an error because the zone is using different DNSSEC policies
 * and should thus use different key directories.
 */

key "keyforview1" {
	algorithm @DEFAULT_HMAC@;
	secret "YPfMoAk6h+3iN8MDRQC004iSNHY=";
};

key "keyforview2" {
	algorithm @DEFAULT_HMAC@;
	secret "4xILSZQnuO1UKubXHkYUsvBRPu8=";
};

options {
	key-directory "keys";
};

view "example1" {
	match-clients { key "keyforview1"; };

	zone "example.net" {
		type primary;
		/* key-directory inherited from options. */
		dnssec-policy "default";
		file "example1.db";
	};
};

view "example2" {
	match-clients { key "keyforview2"; };

	zone "example.net" {
		type primary;
		dnssec-policy "insecure";
		key-directory "keys";
		file "example2.db";
	};
};
