; Copyright (C) Internet Systems Consortium, Inc. ("ISC")
;
; SPDX-License-Identifier: MPL-2.0
;
; This Source Code Form is subject to the terms of the Mozilla Public
; License, v. 2.0.  If a copy of the MPL was not distributed with this
; file, you can obtain one at https://mozilla.org/MPL/2.0/.
;
; See the COPYRIGHT file distributed with this work for additional
; information regarding copyright ownership.



; Use comment lines instead of blank lines to combine update requests into
;	single requests
; Separate update requests for distinct TLDs with blank lines or 'send'
; End the file with a blank line or 'send'

; walled-garden NSIP tests

server 10.53.0.3 @PORT@

; rewrite all of tld2 based on its server IP address
update add  32.2.0.53.10.rpz-nsip.bl.	300 A	    41.41.41.41
update add  32.2.0.53.10.rpz-nsip.bl.	300 AAAA    2041::41
update add  32.2.0.53.10.rpz-nsip.bl.	300 TXT	    "NSIP walled garden"
send

