/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

options {
	query-source address 10.53.0.2;
	notify-source 10.53.0.2;
	transfer-source 10.53.0.2;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { none; };
	recursion yes;
	notify yes;
	serial-query-rate 1; // workaround for KB AA-01213
	dnssec-validation no;
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.2 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

key altkey {
	algorithm @DEFAULT_HMAC@;
	secret "1234abcd8765";
};

view alternate {
	match-clients { key altkey; };

	zone "update.nil" {
		type secondary;
		primaries { 10.53.0.1; };
		file "update.alt.bk";
		allow-transfer { any; };
	};
};

view primary {
	match-clients { any; };

	zone "example.nil" {
		type secondary;
		primaries { 10.53.0.1; };
		file "example.bk";
		allow-transfer { any; };
	};

	zone "update.nil" {
		type secondary;
		primaries { 10.53.0.1; };
		file "update.bk";
		allow-transfer { any; };
	};

	zone "sample" {
		type primary;
		allow-update { any; };
		file "sample.db";
	};
};
