/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

/*
 * DNS64 / RPZ server.
 */

options {
	query-source address 10.53.0.9;
	notify-source 10.53.0.9;
	transfer-source 10.53.0.9;
	port @PORT@;
	pid-file "named.pid";
	statistics-file	"named.stats";
	session-keyfile "session.key";
	listen-on { 10.53.0.9; };
	listen-on-v6 { none; };
	allow-transfer { any; };
	notify yes;
	minimal-responses no;
	recursion yes;
	dnssec-validation no;
	dns64-server "example.localdomain.";
	dns64 64:ff9b::/96 { };
	response-policy {
	    zone "rpz";
	}
	qname-wait-recurse no ;

	include "../dnsrps.conf";
	notify-delay 0;
};


logging { category rpz { default_debug; }; };

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};
controls {
	inet 10.53.0.9 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};


zone "." { type hint; file "hints"; };

zone "rpz." {
	type primary;
	file "rpz.db";
	notify no;
};
