/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

{% if FEATURE_FIPS_DH == "1" %}
include "named-tls.conf";
{% endif %}

options {
	query-source address 10.53.0.2;
	query-source-v6 address fd92:7065:b8e:ffff::2;
	notify-source 10.53.0.2;
	notify-source-v6 fd92:7065:b8e:ffff::2;
	transfer-source 10.53.0.2;
	transfer-source-v6 fd92:7065:b8e:ffff::2;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { fd92:7065:b8e:ffff::2; };
	recursion no;
	dnssec-validation no;

{% if FEATURE_FIPS_DH == "1" %}
	tls-port @TLSPORT@;
	listen-on tls ephemeral { 10.53.0.2; };
	listen-on port @EXTRAPORT1@ tls tls-forward-secrecy { 10.53.0.2; };
	listen-on port @EXTRAPORT2@ tls tls-forward-secrecy-mutual-tls { 10.53.0.2; };
	listen-on port @EXTRAPORT3@ tls tls-expired { 10.53.0.2; };
{% endif %}

};

zone "." {
	type hint;
	file "root.db";
};

zone "example1." {
	type primary;
	file "example.db";
};

zone "example2." {
	type primary;
	file "example.db";
};

zone "example3." {
	type primary;
	file "example.db";
};

zone "example4." {
	type primary;
	file "example.db";
};

zone "example7." {
	type primary;
	file "example.db";
};

zone "example8." {
	type primary;
	file "example.db";
};

zone "example9." {
	type primary;
	file "example.db";
};

zone "grafted." {
	type primary;
	file "example.db";
};

zone "1.0.10.in-addr.arpa." {
	type primary;
	file "example.db";
};

zone "tld" {
	type primary;
	file "tld.db";
};
