/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

options {
	query-source address 10.53.0.1;
	notify-source 10.53.0.1;
	transfer-source 10.53.0.1;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.1; };
	listen-on-v6 { none; };
	recursion no;
	dnssec-validation no;
	notify no;
};

view "in" {
	allow-transfer { any; };

	zone "example." {
		type primary;
		file "example-in.db";
	};

	zone "broken1." {
		type primary;
		file "broken1.db";
	};

	zone "broken2." {
		type primary;
		file "broken2.db";
	};

	zone "broken3." {
		type primary;
		file "broken3.db";
	};

	zone "broken4." {
		type primary;
		file "broken4.db";
	};

	zone "broken5." {
		type primary;
		file "broken5.db";
	};
};

view "class10" class10 {
	zone "." class10 {
		type hint;
		file "class10.hints";
	};

	zone "example." class10 {
		type primary;
		file "example-class10.db";
	};
};
