; Copyright (C) Internet Systems Consortium, Inc. ("ISC")
;
; SPDX-License-Identifier: MPL-2.0
;
; This Source Code Form is subject to the terms of the Mozilla Public
; License, v. 2.0.  If a copy of the MPL was not distributed with this
; file, you can obtain one at https://mozilla.org/MPL/2.0/.
;
; See the COPYRIGHT file distributed with this work for additional
; information regarding copyright ownership.

$ORIGIN .
$TTL 300	; 5 minutes
example			IN SOA	mname1. . (
				1          ; serial
				20         ; refresh (20 seconds)
				20         ; retry (20 seconds)
				1814400    ; expire (3 weeks)
				3600       ; minimum (1 hour)
				)
example.		NS	ns2.example.
ns2.example.		A	10.53.0.2

$ORIGIN example.
a			A	10.0.0.1
a			A	10.0.0.3
a			A	10.0.0.5
			MX	10 mail.example.

mail			A	10.0.0.2
