; Copyright (C) Internet Systems Consortium, Inc. ("ISC")
;
; SPDX-License-Identifier: MPL-2.0
;
; This Source Code Form is subject to the terms of the Mozilla Public
; License, v. 2.0.  If a copy of the MPL was not distributed with this
; file, you can obtain one at https://mozilla.org/MPL/2.0/.
;
; See the COPYRIGHT file distributed with this work for additional
; information regarding copyright ownership.



; Use comment lines instead of blank lines to combine update requests into
;	single requests
; Separate update requests for distinct TLDs with blank lines or 'send'
; End the file with a blank line or 'send'

; the policies or replacements specified in ns3/named.conf override these

server 10.53.0.3 @PORT@

;	1
update add  a3-1.tld2.bl-given.	    300 A 127.0.0.1
send
;	2
update add  a3-2.tld2.bl-passthru.  300 A 127.0.0.2
send
;	3
update add  a3-3.tld2.bl-no-op.	    300 A 127.0.0.3
send
;	4
update add  a3-4.tld2.bl-disabled.  300 A 127.0.0.4
send
;	5 - 7
update add  a3-5.tld2.bl-nodata.    300 A 127.0.0.5
send
;	11
update add  a3-6.tld2.bl-nxdomain.  300 A 127.0.0.11
send
;	12
update add  a3-7.tld2.bl-cname.	    300 A 127.0.0.12
send
;	13
update add  a3-8.tld2.bl-wildcname. 300 A 127.0.0.13
;	14
update add  *.sub9.tld2.bl-wildcname. 300 A 127.0.1.14
send
;	15
update add  a3-15.tld2.bl-garden.   300 A 127.0.0.15
send
;	16
update add  a3-16.tld2.bl.	    300 A 127.0.0.16
send
;	18
update add  a3-18.tld2.bl-drop.	    300 A 127.0.0.18
send
;	19
update add  a3-19.tld2.bl-tcp-only. 300 A 127.0.0.19
send

