/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS9

options {
	port @PORT@;
	pid-file "named.pid";
	listen-on { none; };
	listen-on-v6 { fd92:7065:b8e:ffff::9; };
	recursion yes;
	recursive-clients 0; // regression test for [GL #4987]
	dnssec-validation no;
	dual-stack-servers { fd92:7065:b8e:ffff::7; };
	qname-minimization off;
};


key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet fd92:7065:b8e:ffff::9 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

zone "." {
	type hint;
	file "root.hint";
};
