/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS2

options {
	pid-file "named.pid";
	listen-on port @PORT@ { 10.53.0.2; };
	listen-on-v6 { none; };
	port @PORT@;
	recursion no;
	notify no;
	servfail-ttl 0;
	dnssec-validation no;
	max-records-per-type 2000;
	max-types-per-name 200;
};

zone "example" {
	type primary;
	file "example.db";
};

zone "transfer1" {
	type secondary;
	primaries { 10.53.0.1; };
	file "transfer.db.raw";
};

zone "transfer2" {
	type secondary;
	primaries { 10.53.0.1; };
	masterfile-format text;
	file "transfer.db.txt";
};

zone "transfer3" {
	type secondary;
	primaries { 10.53.0.1; };
	file "formerly-text.db";
};

zone "transfer4" {
	type secondary;
	primaries { 10.53.0.1; };
	masterfile-format text;
	masterfile-style full;
	file "transfer.db.full";
};

zone "under-limit" {
	type secondary;
	primaries { 10.53.0.1; };
	masterfile-format raw;
	file "under-limit.bk";
};

zone "under-limit-kasp" {
	type secondary;
	primaries { 10.53.0.1; };
	masterfile-format raw;
	file "under-limit-kasp.bk";
};

zone "below-limit" {
	type secondary;
	primaries { 10.53.0.1; };
	masterfile-format raw;
	file "below-limit.bk";
};

zone "below-limit-kasp" {
	type secondary;
	primaries { 10.53.0.1; };
	masterfile-format raw;
	file "below-limit-kasp.bk";
};

zone "255types" {
	type secondary;
	primaries { 10.53.0.1; };
	masterfile-format raw;
	file "255types.bk";
};
