/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

options {
	query-source address 10.53.0.2;
	notify-source 10.53.0.2;
	transfer-source 10.53.0.2;
	port @PORT@;
	session-keyfile "session.key";
	pid-file "named.pid";
	statistics-file	"named.stats";
	listen-on { 10.53.0.2; };
	listen-on-v6 { none; };
	notify no;
	recursion yes;
	dnssec-validation no;

	rate-limit {
	    responses-per-second 2;
	    all-per-second 50;
	    slip 3;
	    exempt-clients { 10.53.0.7; };

	    // small enough to force a table expansion
	    min-table-size 75;
	};
};


key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};
controls {
	inet 10.53.0.2 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

/*
 * These log settings have no effect unless "-g" is removed from ../../start.pl
 */
logging {
	channel debug {
	    file "log-debug";
	    print-category yes; print-severity yes; severity debug 10;
	};
	channel queries {
	    file "log-queries";
	    print-category yes; print-severity yes; severity info;
	};
	category rate-limit { debug; queries; };
	category queries { debug; queries; };
};

zone "." { type hint; file "hints"; };

zone "tld2."{ type primary; file "tld2.db"; };
