#!/usr/bin/perl

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

# server-xml.pl:
# Parses the XML version of the server stats into a normalized format.

use XML::Simple;

my $ref = XMLin("xml.stats");
my $counters = $ref->{server}->{counters};
foreach $group (@$counters) {
    foreach $key (keys %{$group->{counter}}) {
        print $group->{type} . " " . $key . ": ". $group->{counter}->{$key}->{content} . "\n";
    }
}
