/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS2

options {
	query-source address 10.53.0.2;
	notify-source 10.53.0.2;
	transfer-source 10.53.0.2;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { fd92:7065:b8e:ffff::2; };
	recursion no;
	dnssec-validation no;
	geoip-directory "../data";
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.2 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

view one {
	match-clients { geoip org "One Systems, Inc."; };
	zone "example" {
		type primary;
		file "example1.db";
	};
};

view two {
	match-clients { geoip org "Two Technology Ltd."; };
	zone "example" {
		type primary;
		file "example2.db";
	};
};

view three {
	match-clients { geoip org "Three Network Labs"; };
	zone "example" {
		type primary;
		file "example3.db";
	};
};

view four {
	match-clients { geoip org "Four University"; };
	zone "example" {
		type primary;
		file "example4.db";
	};
};

view five {
	match-clients { geoip org "Five Telecom"; };
	zone "example" {
		type primary;
		file "example5.db";
	};
};

view six {
	match-clients { geoip org "Six Company"; };
	zone "example" {
		type primary;
		file "example6.db";
	};
};

view seven {
	match-clients { geoip org "Seven Communications"; };
	zone "example" {
		type primary;
		file "example7.db";
	};
};

view none {
	match-clients { any; };
	zone "example" {
		type primary;
		file "example.db.in";
	};
};
