/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

key "a" {
	algorithm @DEFAULT_HMAC@;
	secret "aaaaaaaaaaaaaaaaaaaa";
};

key "b" {
	algorithm @DEFAULT_HMAC@;
	secret "bbbbbbbbbbbbbbbbbbbb";
};

key "c" {
	algorithm @DEFAULT_HMAC@;
	secret "cccccccccccccccccccc";
};

options {
	query-source address 10.53.0.5;
	notify-source 10.53.0.5;
	transfer-source 10.53.0.5;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.5; };
	listen-on-v6 { none; };
	allow-transfer { any; };
	recursion yes;
	notify yes;
	dnssec-validation no;
};

view "a" {
	match-clients { key "a"; };
	zone "x21" {
		type primary;
		also-notify { 10.53.0.5 key "b"; 10.53.0.5 key "c"; };
		file "x21.db";
		allow-update { any; };
	};
};

view "b" {
	match-clients { key "b"; };
	zone "x21" {
		type secondary;
		primaries { 10.53.0.5 key "a"; };
		file "x21.bk-b";
		notify no;
	};
};

view "c" {
	match-clients { key "c"; };
	zone "x21" {
		type secondary;
		primaries { 10.53.0.5 key "a"; };
		file "x21.bk-c";
		notify no;
	};
};
