/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS2

acl rfc1918 { 10/8; 192.168/16; 172.16/12; };

options {
	query-source address 10.53.0.2;
	notify-source 10.53.0.2;
	transfer-source 10.53.0.2;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { none; };
	recursion yes;
	notify yes;
	dnssec-validation no;

	dns64 2001:aaaa::/96 {
		clients { 10.53.0.2; };
		mapped { !rfc1918; any; };
		exclude { 2001:eeee::/32; 64:FF9B::/96; ::ffff:0000:0000/96; };
		suffix ::;
	};

	dns64 64:FF9B::/96 {
		clients { 10.53.0.1; };
		mapped { !192.228.79.201; !rfc1918; any; };
		exclude { 64:FF9B::/96; ::ffff:0000:0000/96; };
		suffix ::;
	};

	dns64 2001:bbbb::/96 {
		clients { 10.53.0.4; };
		mapped { !rfc1918; any; };
		suffix ::;
	};

	dns64-server "dns64.example.net.";
	dns64-contact "hostmaster.example.net.";
	dns64 2001:32::/32 { clients { 10.53.0.6; }; };
	dns64 2001:40::/40 { clients { 10.53.0.6; }; };
	dns64 2001:48::/48 { clients { 10.53.0.6; }; };
	dns64 2001:56::/56 { clients { 10.53.0.6; }; };
	dns64 2001:64::/64 { clients { 10.53.0.6; }; };

	dns64 2001:96::/96 { clients { 10.53.0.7; }; };

	response-policy { zone "rpz"; };
};


zone "." {
	type hint;
	file "../../_common/root.hint";
};

zone "rpz" {
	type primary;
	file "rpz.db";
};
