/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

options {
	query-source address 10.53.0.1;
	notify-source 10.53.0.1;
	transfer-source 10.53.0.1;
	port @PORT@;
	session-keyfile "session.key";
	pid-file "named.pid";
	listen-on { 10.53.0.1; };
	listen-on-v6 { none; };
	dnssec-validation auto;
	bindkeys-file "../../../../../bind.keys";
	minimal-responses no;
	recursion no;
	notify yes;
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.1 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

zone . {
	type hint;
	file "../../_common/root.hint.blackhole";
};

zone changed {
	type primary;
	update-policy local;
	file "changed.db";
};

zone unchanged {
	type primary;
	update-policy local;
	file "unchanged.db";
};

zone changed2 {
	type primary;
	update-policy local;
	file "changed2.db";
};

zone unchanged2 {
	type primary;
	update-policy local;
	file "unchanged2.db";
};

zone hdr1d1d2d1d2 {
	type primary;
	update-policy local;
	file "d1212.db";
};

zone hdr1d2d1d2d1 {
	type primary;
	update-policy local;
	file "d2121.db";
};

zone ixfr {
	type primary;
	ixfr-from-differences yes;
	file "ixfr.db";
};

zone maxjournal {
	type primary;
	max-journal-size 1k;
	update-policy local;
	file "maxjournal.db";
};

zone maxjournal2 {
	type primary;
	max-journal-size 1k;
	update-policy local;
	file "maxjournal2.db";
};
