/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS1

options {
	pid-file "named.pid";
	listen-on port @PORT@ { 10.53.0.1; };
	port @PORT@;
	listen-on-v6 { none; };
	recursion no;
	notify no;
	session-keyfile "session.key";
	servfail-ttl 0;
	dnssec-validation no;
	max-records-per-type 2050;
	max-types-per-name 500;
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.1 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

dnssec-policy "masterformat" {
	keys {
		ksk key-directory lifetime unlimited algorithm @DEFAULT_ALGORITHM@;
		zsk key-directory lifetime unlimited algorithm @DEFAULT_ALGORITHM@;
	};
};

zone "example" {
	type primary;
	masterfile-format raw;
	file "example.db.raw";
};

zone "compat-example" {
	type primary;
	masterfile-format raw;
	file "example.db.compat";
};

zone "transfer1" {
	type primary;
	file "example.db";
	allow-transfer { any; };
};

zone "transfer2" {
	type primary;
	file "example.db";
	allow-transfer { any; };
};

zone "transfer3" {
	type primary;
	file "example.db";
	allow-transfer { any; };
};

zone "transfer4" {
	type primary;
	file "example.db";
	allow-transfer { any; };
};


zone "under-limit" {
	type primary;
	file "under-limit.db.raw";
	masterfile-format raw;
	allow-transfer { any; };
};

zone "under-limit-kasp" {
	type primary;
	file "under-limit-kasp.db.raw";
	masterfile-format raw;
	dnssec-policy masterformat;
	allow-transfer { any; };
};

zone "below-limit" {
	type primary;
	file "below-limit.db.raw";
	masterfile-format raw;
	allow-transfer { any; };
};

zone "below-limit-kasp" {
	type primary;
	file "below-limit-kasp.db.raw";
	masterfile-format raw;
	dnssec-policy masterformat;
	inline-signing no;
	allow-update { any; };
	allow-transfer { any; };
};

zone "above-limit" {
	type primary;
	file "above-limit.db.raw";
	masterfile-format raw;
	allow-transfer { any; };
};

zone "255types" {
	type primary;
	file "255types.db.raw";
	masterfile-format raw;
	allow-transfer { any; };
};

zone "signed" {
	type primary;
	file "signed.db.raw";
	masterfile-format raw;
	allow-transfer { any; };
	update-policy local;
	dnssec-policy masterformat;
};
