/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

include "../../_common/rndc.key";

controls {
	inet 10.53.0.4 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

options {
	query-source address 10.53.0.4;
	notify-source 10.53.0.4;
	transfer-source 10.53.0.4;
	port @PORT@;
	tls-port @TLSPORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.4; };
	listen-on-v6 { fd92:7065:b8e:ffff::4; };
	allow-transfer { any; };
	notify no;
	notify-delay 0;
	recursion no;
	serial-query-rate 100;
	ixfr-from-differences yes; // GL #3777
	dnssec-validation no;

	catalog-zones {
		zone "catalog-misc.example"
			min-update-interval 1s
			default-primaries { 10.53.0.1; };
		zone "catalog-tls.example"
			min-update-interval 1s
			default-primaries { 10.53.0.1 key tsig_key tls ephemeral; };
		zone "catalog-self.example"
			min-update-interval 1s
			default-primaries { 10.53.0.4; };
	};
};

zone "catalog-tls.example" {
	type secondary;
	file "catalog-tls.example.db";
	primaries { 10.53.0.1 key tsig_key tls ephemeral; };
};

zone "catalog-misc.example" {
	type secondary;
	file "catalog-misc.example.db";
	primaries { 10.53.0.1; };
};

zone "catalog-self.example" {
	type primary;
	file "catalog-self.example.db";
	notify explicit;
};

key tsig_key. {
	secret "LSAnCU+Z";
	algorithm @DEFAULT_HMAC@;
};

key next_key. {
	secret "LaAnCU+Z";
	algorithm @DEFAULT_HMAC@;
};

key longlonglongname0123456789abcdef. {
	secret "LaAnCU+Z";
	algorithm @DEFAULT_HMAC@;
};
