/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

options {
	query-source address 10.53.0.1;
	notify-source 10.53.0.1;
	transfer-source 10.53.0.1;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.1; };
	listen-on-v6 { none; };
	recursion no;
	dnssec-validation no;
	notify yes;

	sortlist {
	    { 10.53.0.1;				   // IF 10.53.0.1
		{
		    !1.1.1.4; !1.1.1.2; !1.1.1.3; !1.1.1.1; // sort these last,
		    192.168.3/24;			    // this first
		    { 192.168.2/24; 192.168.1/24; }; }; };  // and these next
	    { { 10.53.0.2; 10.53.0.3; }; };		    // Prefer self
	    10.53.0.4;					    // BIND 8 compat
	    { 10.53.0.5; 10.53.0.5; };    		    // BIND 8 compat
	};
};

zone "." {
	type primary;
	file "root.db";
};

zone "example" {
	type primary;
	file "example.db";
};
