/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

{% set csk_roll = csk_roll | default(False) %}
{% set _csk_file = "csk1.conf" if not csk_roll else "csk2.conf" %}
{% set zones = ["kasp", "manual"] %}

include "@_csk_file@";
include "named.common.conf";

{% for tld in zones %}
zone "step1.csk-algorithm-roll.@tld@" {
	type primary;
	file "step1.csk-algorithm-roll.@tld@.db";
	dnssec-policy "csk-algoroll-@tld@";
};

{% if csk_roll %}
zone "step2.csk-algorithm-roll.@tld@" {
	type primary;
	file "step2.csk-algorithm-roll.@tld@.db";
	dnssec-policy "csk-algoroll-@tld@";
};

zone "step3.csk-algorithm-roll.@tld@" {
	type primary;
	file "step3.csk-algorithm-roll.@tld@.db";
	dnssec-policy "csk-algoroll-@tld@";
};

zone "step4.csk-algorithm-roll.@tld@" {
	type primary;
	file "step4.csk-algorithm-roll.@tld@.db";
	dnssec-policy "csk-algoroll-@tld@";
};

zone "step5.csk-algorithm-roll.@tld@" {
	type primary;
	file "step5.csk-algorithm-roll.@tld@.db";
	dnssec-policy "csk-algoroll-@tld@";
};

zone "step6.csk-algorithm-roll.@tld@" {
	type primary;
	file "step6.csk-algorithm-roll.@tld@.db";
	dnssec-policy "csk-algoroll-@tld@";
};
{% endif %}
{% endfor %}
