/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS2

options {
	query-source address 10.53.0.2;
	notify-source 10.53.0.2;
	transfer-source 10.53.0.2;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.2; };
	listen-on-v6 { none; };
	recursion no;
	dnssec-validation no;
	notify yes;
};

zone . {
       type hint;
       file "../../_common/root.hint";
};

zone "example" {
	type primary;
	file "example.db.signed";
	allow-update { any; };
};

zone "sub2.example" {
	type primary;
	file "sub.db";
};

zone "signed-sub2.example" {
	type primary;
	file "sub.db";
};

zone "wildcard-secure.example" {
	type primary;
	file "wildcard-secure.example.db.signed";
};

zone "wildcard-nsec.example" {
	type primary;
	file "wildcard-nsec.example.db.signed";
};

zone "wildcard-nsec3.example" {
	type primary;
	file "wildcard-nsec3.example.db.signed";
};

zone "wildcard-nsec3-optout.example" {
	type primary;
	file "wildcard-nsec3-optout.example.db.signed";
};

zone "domain0.nil" { type primary; file "generic.db"; };
zone "domain1.nil" { type primary; file "generic.db"; };
zone "domain2.nil" { type primary; file "generic.db"; };
zone "domain3.nil" { type primary; file "generic.db"; };
zone "domain4.nil" { type primary; file "generic.db"; };
zone "domain5.nil" { type primary; file "generic.db"; };
zone "domain6.nil" { type primary; file "generic.db"; };
zone "domain7.nil" { type primary; file "generic.db"; };
zone "domain8.nil" { type primary; file "generic.db"; };
zone "domain9.nil" { type primary; file "generic.db"; };
