/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

# dnsrpzd configuration.

pid-file ../dnsrpzd.pid

# configure NOTIFY and zone transfers
port @EXTRAPORT1@;
listen-on port @EXTRAPORT1@ { 10.53.0.3; };
allow-notify { 10.53.0.0/24; };

zone "bl0"		{type primary; file "../ns5/bl.db"; };
zone "bl1"		{type primary; file "../ns5/bl.db"; };
zone "bl2"		{type primary; file "../ns5/bl.db"; };
zone "bl3"		{type primary; file "../ns5/bl.db"; };
zone "bl4"		{type primary; file "../ns5/bl.db"; };
zone "bl5"		{type primary; file "../ns5/bl.db"; };
zone "bl6"		{type primary; file "../ns5/bl.db"; };
zone "bl7"		{type primary; file "../ns5/bl.db"; };
zone "bl8"		{type primary; file "../ns5/bl.db"; };
zone "bl9"		{type primary; file "../ns5/bl.db"; };
zone "bl10"		{type primary; file "../ns5/bl.db"; };
zone "bl11"		{type primary; file "../ns5/bl.db"; };
zone "bl12"		{type primary; file "../ns5/bl.db"; };
zone "bl13"		{type primary; file "../ns5/bl.db"; };
zone "bl14"		{type primary; file "../ns5/bl.db"; };
zone "bl15"		{type primary; file "../ns5/bl.db"; };
zone "bl16"		{type primary; file "../ns5/bl.db"; };
zone "bl17"		{type primary; file "../ns5/bl.db"; };
zone "bl18"		{type primary; file "../ns5/bl.db"; };
zone "bl19"		{type primary; file "../ns5/bl.db"; };

zone "bl"		{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-2"		{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-given"		{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-passthru"	{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-no-op"		{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-disabled"	{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-nodata"	{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-nxdomain"	{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-cname"		{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-wildcname"	{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-garden"	{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-drop"		{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl-tcp-only"	{type secondary; primaries port @PORT@ { 10.53.0.3; }; };
zone "bl.tld2"		{type secondary; primaries port @PORT@ { 10.53.0.3; }; };

zone "policy1"		{type secondary; primaries port @PORT@ { 10.53.0.6; }; };
zone "policy2"		{type secondary; primaries port @PORT@ { 10.53.0.7; }; };
