/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

/*
 * Test serve-stale interaction with fetch-limits (dual-mode).
 */

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.3 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

options {
	query-source address 10.53.0.3;
	notify-source 10.53.0.3;
	transfer-source 10.53.0.3;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.3; };
	listen-on-v6 { none; };
	dnssec-validation no;
	recursion yes;
	/*
	 * stale-answer-enable is not strictly required because serving
	 * stale answers is enabled in the test via rndc.
	 */
	stale-answer-enable yes;
	stale-cache-enable yes;
	stale-answer-ttl 3;
	stale-refresh-time 4;
	resolver-query-timeout 10000; # 10 seconds
	fetches-per-zone 1 fail;
	fetches-per-server 1 fail;
	max-stale-ttl 3600;
};

zone "." {
	type secondary;
	primaries { 10.53.0.1; };
	file "root.bk";
};
