/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS4

options {
	query-source address 10.53.0.4;
	notify-source 10.53.0.4;
	transfer-source 10.53.0.4;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.4; };
	listen-on-v6 { none; };
	allow-transfer { any; };
	recursion no;
	key-directory ".";
	dnssec-validation no;
};

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.4 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

dnssec-policy "rsasha256" {
	keys {
		zsk key-directory lifetime P3M algorithm 8 2048;
		ksk key-directory lifetime P1Y algorithm 8 2048;
	};

	dnskey-ttl 300;
	publish-safety 1h;
	retire-safety 1h;

	signatures-refresh 5d;
	signatures-validity 14d;
	signatures-validity-dnskey 14d;

	max-zone-ttl 1d;
	zone-propagation-delay 300;

	parent-ds-ttl 86400;
	parent-propagation-delay 3h;
};

key "external" {
	algorithm @DEFAULT_HMAC@;
	secret "YPfMoAk6h+3iN8MDRQC004iSNHY=";
};

key "internal" {
	algorithm @DEFAULT_HMAC@;
	secret "4xILSZQnuO1UKubXHkYUsvBRPu8=";
};

view "ext" {
        match-clients { key "external"; };

	zone "view-rsasha256.kasp" {
		type primary;
		file "view-rsasha256.kasp.ext.db";
		dnssec-policy "rsasha256";
		inline-signing no;
		allow-update { any; };
	};
};

view "int" {
        match-clients { key "internal"; };

	zone "view-rsasha256.kasp" {
		type primary;
		file "view-rsasha256.kasp.int.db";
		dnssec-policy "rsasha256";
		inline-signing no;
		allow-update { any; };
	};
};
