/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

include "kasp.conf";
include "named.common.conf";

/* RFC 8901 Multi-signer Model 2. */
zone "multisigner-model2.kasp" {
        type primary;
        file "multisigner-model2.kasp.db";
        dnssec-policy "multisigner-model2";
        allow-update { any; };
};

/*
 * A zone that starts with keys that have tags that are
 * outside of the desired multi-signer key tag range.
 */
zone "single-to-multisigner.kasp" {
        type primary;
        file "single-to-multisigner.kasp.db";
        dnssec-policy "multisigner-model2";
        allow-update { any; };
};
