/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */


// NS3

acl rfc1918 { 10/8; 192.168/16; 172.16/12; };

options {
	query-source address 10.53.0.3;
	query-source-v6 address fd92:7065:b8e:ffff::3;
	notify-source 10.53.0.3;
	notify-source-v6 fd92:7065:b8e:ffff::3;
	transfer-source 10.53.0.3;
	transfer-source-v6 fd92:7065:b8e:ffff::3;
	port @EXTRAPORT1@;
	pid-file "named.pid";
	listen-on port @PORT@ { 10.53.0.3; };	// for start.pl
	listen-on { 10.53.0.3; };
	listen-on-v6 { fd92:7065:b8e:ffff::3; };
	notify yes;
	dnssec-validation no;
	allow-recursion { any; };
	resolver-use-dns64 yes;

	dns64 fd92:7065:b8e:fffe::/96 {
		clients { any; };
		mapped { 10.53.0.4; !rfc1918; any; };
		suffix ::;
	};
};


zone "." {
	type hint;
	file "hints";
};
