/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

tls tls-forward-secrecy {
    protocols { TLSv1.2; };
    ciphers "HIGH:!kRSA:!aNULL:!eNULL:!RC4:!3DES:!MD5:!EXP:!PSK:!SRP:!DSS:!SHA1:!SHA256:!SHA384";
    prefer-server-ciphers yes;
    key-file "../CA/certs/srv02.crt01.example.com.key";
    cert-file "../CA/certs/srv02.crt01.example.com.pem";
    dhparam-file "../dhparam3072.pem";
};

tls tls-forward-secrecy-mutual-tls {
    protocols { TLSv1.2; };
    ciphers "HIGH:!kRSA:!aNULL:!eNULL:!RC4:!3DES:!MD5:!EXP:!PSK:!SRP:!DSS:!SHA1:!SHA256:!SHA384";
    prefer-server-ciphers yes;
    key-file "../CA/certs/srv02.crt01.example.com.key";
    cert-file "../CA/certs/srv02.crt01.example.com.pem";
    dhparam-file "../dhparam3072.pem";
    ca-file "../CA/CA.pem";
};

tls tls-expired {
    protocols { TLSv1.2; };
    ciphers "HIGH:!kRSA:!aNULL:!eNULL:!RC4:!3DES:!MD5:!EXP:!PSK:!SRP:!DSS:!SHA1:!SHA256:!SHA384";
    prefer-server-ciphers yes;
    key-file "../CA/certs/srv02.crt01-expired.example.com.key";
    cert-file "../CA/certs/srv02.crt01-expired.example.com.pem";
    dhparam-file "../dhparam3072.pem";
};
