/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

options {
	query-source address 10.53.0.7;
	notify-source 10.53.0.7;
	transfer-source 10.53.0.7;
	port @PORT@;
	pid-file "named.pid";
	statistics-file "named.stats";
	session-keyfile "session.key";
	listen-on { 10.53.0.7; };
	listen-on-v6 { none; };
	allow-transfer { any; };
	minimal-responses no;
	recursion yes;
	dnssec-validation yes;

	response-policy {
	    zone "policy2" add-soa no ede none;
	} qname-wait-recurse no
	nsip-enable yes
	nsdname-enable yes
	min-update-interval 0;

	include "../dnsrps.conf";
};

logging { category rpz { default_debug; }; };

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.7 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

include "../trusted.conf";

zone "policy2" {
	type secondary;
	primaries { 10.53.0.5; };
	file "policy2.db";
	also-notify { 10.53.0.3 port @EXTRAPORT1@; };
	notify-delay 0;
	allow-transfer	{ any; };
	request-ixfr no; // force axfr on rndc reload
};
