/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

include "../../_common/rndc.key";

controls {
	inet 10.53.0.1 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

tls tls-perfect-forward-secrecy {
	protocols { TLSv1.3; };
	cipher-suites "TLS_AES_256_GCM_SHA384:TLS_CHACHA20_POLY1305_SHA256:TLS_AES_128_GCM_SHA256";
	key-file "../self-signed-key.pem";
	cert-file "../self-signed-cert.pem";
	session-tickets no;
};

tls tls-pfs-aes256 {
	protocols { TLSv1.3; };
	cipher-suites "TLS_AES_256_GCM_SHA384";
	key-file "../self-signed-key.pem";
	cert-file "../self-signed-cert.pem";
	session-tickets no;
};

tls tls-pfs-aes128 {
	protocols { TLSv1.3; };
	cipher-suites "TLS_AES_128_GCM_SHA256";
	key-file "../self-signed-key.pem";
	cert-file "../self-signed-cert.pem";
	session-tickets no;
};

tls tls-pfs-chacha20 {
	protocols { TLSv1.3; };
	cipher-suites "TLS_CHACHA20_POLY1305_SHA256";
	key-file "../self-signed-key.pem";
	cert-file "../self-signed-cert.pem";
	session-tickets no;
};

options {
	port @PORT@;
	tls-port @TLSPORT@;
	pid-file "named.pid";
	listen-on-v6 { none; };
	listen-on { 10.53.0.1; };
	listen-on tls tls-perfect-forward-secrecy { 10.53.0.1; };
	listen-on port @EXTRAPORT1@ tls tls-pfs-aes128 { 10.53.0.1; };
	listen-on port @EXTRAPORT2@ tls tls-pfs-aes256 { 10.53.0.1; };
	listen-on port @EXTRAPORT3@ tls tls-pfs-chacha20 { 10.53.0.1; };
	recursion no;
	notify explicit;
	also-notify { 10.53.0.2 port @PORT@; };
	statistics-file "named.stats";
	dnssec-validation no;
	tcp-initial-timeout 1200;
	transfers-in 100;
	transfers-out 100;
};


zone "." {
	type primary;
	file "root.db";
	allow-transfer port @TLSPORT@ transport tls { any; };
};

zone "example" {
	type primary;
	file "example.db";
	allow-transfer port @TLSPORT@ transport tls { any; };
};

zone "example-aes-128" {
	type primary;
	file "example.db";
	allow-transfer port @EXTRAPORT1@ transport tls { any; };
};

zone "example-aes-256" {
	type primary;
	file "example.db";
	allow-transfer port @EXTRAPORT2@ transport tls { any; };
};

zone "example-chacha-20" {
	type primary;
	file "example.db";
	allow-transfer port @EXTRAPORT3@ transport tls { any; };
};
