/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

dnssec-policy "csk-algoroll" {
	signatures-refresh P5D;
	signatures-validity 30d;
	signatures-validity-dnskey 30d;

	keys {
		csk lifetime unlimited algorithm rsasha256;
	};

	dnskey-ttl 1h;
	publish-safety PT1H;
	retire-safety 2h;
	zone-propagation-delay 3600;
	max-zone-ttl 6h;
	parent-propagation-delay pt1h;
	parent-ds-ttl 7200;
};
