; Copyright (C) Internet Systems Consortium, Inc. ("ISC")
;
; SPDX-License-Identifier: MPL-2.0
;
; This Source Code Form is subject to the terms of the Mozilla Public
; License, v. 2.0.  If a copy of the MPL was not distributed with this
; file, you can obtain one at https://mozilla.org/MPL/2.0/.
;
; See the COPYRIGHT file distributed with this work for additional
; information regarding copyright ownership.




; Use comment lines instead of blank lines to combine update requests into
;	single requests
; Separate update requests for distinct TLDs with blank lines or 'send'
; End the file with a blank line or 'send'
; CNAME targets are absolute even without trailing "."

; IP tests

server 10.53.0.3 @PORT@

; NODATA a3-1.tld2
;	1
update add  32.1.3.168.192.rpz-ip.bl	300 CNAME *.
;
; NXDOMAIN for 192.168.4.0/24, the network of a4-1.tld2 and a4-2.tld2
;	4
update add  24.0.4.168.192.rpz-ip.bl	300 CNAME .
;
; old passthru in NXDOMAIN CIDR block to leave a4-1.tld2 unchanged
;	3
update add  32.1.4.168.192.rpz-ip.bl	300 CNAME 32.1.4.168.192
;
; NODATA for a4-3.tld2
;	8
update add  32.3.4.168.192.rpz-ip.bl	300 CNAME *.
;
; NXDOMAIN for IPv6 a3-1.tld2
;	9
update add  128.1.zz.3.2.2001.rpz-ip.bl	300 CNAME .
;
; apply the policy with the lexically smaller trigger address of 192.168.5.1
; to an RRset of more than one A RR
;	11
update add  32.1.5.168.192.rpz-ip.bl	300 A	127.0.0.1
update add  32.2.5.168.192.rpz-ip.bl	300 A	127.0.0.2
;
; prefer first conflicting IP zone for a5-3.tld2
;	12
update add 32.3.5.168.192.rpz-ip.bl	300 A	127.0.0.1
; non canonical form entry to trigger log message
update add  128.2.0.0.0.0.3.2.2001.rpz-ip.bl	300 CNAME .
send

;
update add 32.3.5.168.192.rpz-ip.bl-2	300 A	127.0.0.2
send

; prefer QNAME to IP for a5-4.tld2
;	13, 14
update add 32.4.5.168.192.rpz-ip.bl	300 CNAME a12.tld2.
update add a5-4.tld2.bl			300 CNAME a14.tld4.
;
; poke hole in NXDOMAIN CIDR block to leave a4-4.tld2 unchanged
;	15
update add  32.4.4.168.192.rpz-ip.bl	300 CNAME rpz-passthru.
;
; assert in rbtdb.c
;	16
update add 32.16.1.16.172.rpz-ip.bl	300 CNAME .
send
update add c2.crash2.tld3.bl-2		300 A	127.0.0.16
send

; client-IP address trigger
;	17
update add 32.1.0.53.10.rpz-client-ip.bl 300 A	127.0.0.17
send

