/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.3 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

options {
	query-source address 10.53.0.3;
	notify-source 10.53.0.3;
	transfer-source 10.53.0.3;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.3; };
	listen-on-v6 { none; };
	recursion yes;
	allow-query-cache { 10.53.0.1; };
	trust-anchor-telemetry yes;
	allow-new-zones yes;
	dnssec-validation yes;
};

zone "." {
	type hint;
	file "../../_common/root.hint";
};

zone "." {
	type mirror;
	primaries { 10.53.0.1; };
	file "root.db.mirror";
};

zone "initially-unavailable" {
	type mirror;
	primaries { 10.53.0.2; };
	file "initially-unavailable.db.mirror";
};

zone "verify-axfr" {
	type mirror;
	primaries { 10.53.0.2; };
	file "verify-axfr.db.mirror";
};

zone "verify-csk" {
	type mirror;
	primaries { 10.53.0.2; };
	file "verify-csk.db.mirror";
};

zone "verify-ixfr" {
	type mirror;
	primaries { 10.53.0.2; };
	file "verify-ixfr.db.mirror";
	masterfile-format text;
};

zone "verify-load" {
	type mirror;
	primaries { 10.53.0.2; };
	file "verify-load.db.mirror";
	masterfile-format text;
};

zone "verify-reconfig" {
	type mirror;
	primaries { 10.53.0.2; };
	file "verify-reconfig.db.mirror";
	masterfile-format text;
};

zone "verify-unsigned" {
	type mirror;
	primaries { 10.53.0.2; };
	file "verify-unsigned.db.mirror";
};

zone "verify-untrusted" {
	type mirror;
	primaries { 10.53.0.2; };
	file "verify-untrusted.db.mirror";
};

include "../ns1/trusted.conf";
include "../ns2/trusted-mirror.conf";
