/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

options {
	query-source address 10.53.0.1;
	notify-source 10.53.0.1;
	transfer-source 10.53.0.1;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.1; };
	listen-on-v6 { none; };
	recursion no;
	dnssec-validation no;
	notify yes;
};

zone "." { type primary; file "root.db.signed"; };

/*
 * RFC 4592 example zone.
 */
zone "allwild.test" { type primary; file "allwild.db"; };
zone "example" { type primary; file "example.db"; };
zone "nsec" { type primary; file "nsec.db.signed"; };
zone "private.nsec" { type primary; file "private.nsec.db.signed"; };

zone "nestedwild.test" {
        type primary;
        file "nestedwild.db";
        check-names ignore;
};

/*
 * The contents of nsec3 and private.nsec3 are specially chosen to
 * have separate NSEC3 records for the "no qname proof" and the
 * "closest encloser proof".
 */
zone "nsec3" { type primary; file "nsec3.db.signed"; };
zone "private.nsec3" { type primary; file "private.nsec3.db.signed"; };
