/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

plugin query "../../../plugins/.libs/filter-aaaa.so" {
	/*
	 * While this matches the defaults, it is not a good configuration
	 * to have in named.conf as the two options contradict each other
	 * indicating a error on behalf of the operator.
	 *
	 * The default is to have filter-aaaa-on-v4 off, but if it is turned
	 * on then it applies to all IPv4 queries.  This results in
	 * contradictory defaults.
	 */
	filter-aaaa-on-v4 no;
	filter-aaaa { any; };
};
