/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

options {
	query-source address 10.53.0.1;
	notify-source 10.53.0.1;
	transfer-source 10.53.0.1;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.1; };
	listen-on-v6 { none; };
	recursion yes;
	dnssec-validation no;
	notify no;
};

{% if FEATURE_MD5 == "1" %}
include "named-md5.conf";
{% endif %}

key "sha1" {
	secret "FrSt77yPTFx6hTs4i2tKLB9LmE0=";
	algorithm hmac-sha1;
};

key "sha224" {
	secret "hXfwwwiag2QGqblopofai9NuW28q/1rH4CaTnA==";
	algorithm hmac-sha224;
};

key "sha256" {
	secret "R16NojROxtxH/xbDl//ehDsHm5DjWTQ2YXV+hGC2iBY=";
	algorithm hmac-sha256;
};

key "sha384" {
	secret "OaDdoAk2LAcLtYeUnsT7A9XHjsb6ZEma7OCvUpMraQIJX6HetGrlKmF7yglO1G2h";
	algorithm hmac-sha384;
};

key "sha512" {
	secret "jI/Pa4qRu96t76Pns5Z/Ndxbn3QCkwcxLOgt9vgvnJw5wqTRvNyk3FtD6yIMd1dWVlqZ+Y4fe6Uasc0ckctEmg==";
	algorithm hmac-sha512;
};

key "hmac-sha1-legacy" {
	algorithm "hmac-sha1";
	secret "TxGx9XBp6Pp5yYAOKXdERA==";
};

key "hmac-sha224-legacy" {
	algorithm "hmac-sha224";
	secret "H8Hyw718rLqToQFRLAeFWQ==";
};

key "hmac-sha256-legacy" {
	algorithm "hmac-sha256";
	secret "fdT9hiPov4ThMEfRv1FNmA==";
};

key "hmac-sha384-legacy" {
	algorithm "hmac-sha384";
	secret "fnshFIjQTLFap6+j2JGBkA==";
};

key "hmac-sha512-legacy" {
	algorithm "hmac-sha512";
	secret "BZwNLICp2tj4hi6gil41eg==";
};

key "sha1-trunc" {
	secret "FrSt77yPTFx6hTs4i2tKLB9LmE0=";
	algorithm hmac-sha1-80;
};

key "sha224-trunc" {
	secret "hXfwwwiag2QGqblopofai9NuW28q/1rH4CaTnA==";
	algorithm hmac-sha224-112;
};

key "sha256-trunc" {
	secret "R16NojROxtxH/xbDl//ehDsHm5DjWTQ2YXV+hGC2iBY=";
	algorithm hmac-sha256-128;
};

key "sha384-trunc" {
	secret "OaDdoAk2LAcLtYeUnsT7A9XHjsb6ZEma7OCvUpMraQIJX6HetGrlKmF7yglO1G2h";
	algorithm hmac-sha384-192;
};

key "sha512-trunc" {
	secret "jI/Pa4qRu96t76Pns5Z/Ndxbn3QCkwcxLOgt9vgvnJw5wqTRvNyk3FtD6yIMd1dWVlqZ+Y4fe6Uasc0ckctEmg==";
	algorithm hmac-sha512-256;
};

zone "example.nil" {
	type primary;
	file "example.db";
};

server 10.53.0.2 {
	keys sha256;
};

zone "bad-tsig" {
	type forward;
	forwarders { 10.53.0.2; };
	forward only;
};
