/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS7

options {
	query-source address 10.53.0.7;
	notify-source 10.53.0.7;
	transfer-source 10.53.0.7;
	port @PORT@;
	pid-file "named.pid";
	listen-on { 10.53.0.7; };
	listen-on-v6 { fd92:7065:b8e:ffff::7; };
	recursion yes;
	dnssec-validation no;
	empty-zones-enable yes;
	disable-empty-zone 20.172.in-addr.arpa;
	/*
	 * check prefetch disabled
	 * check zero ttl not returned
	 */
	prefetch 0;
	querylog yes;
	edns-udp-size 4096;
};


key rndc_key {
	secret "1234abcd8765";
	algorithm @DEFAULT_HMAC@;
};

controls {
	inet 10.53.0.7 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

zone "." {
	type hint;
	file "root.hint";
};

zone "server" {
	type primary;
	file "server.db";
	allow-update { any; };
};

zone "edns-version.tld" {
	type primary;
	file "edns-version.tld.db";
};

zone "all-cnames" {
	type primary;
	file "all-cnames.db";
};

zone "tld2" {
	type primary;
	file "tld2.db";
};

zone "sub.tld1" {
	type primary;
	file "sub.tld1.db";
};
