/*
 * Copyright (C) Internet Systems Consortium, Inc. ("ISC")
 *
 * SPDX-License-Identifier: MPL-2.0
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0.  If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * See the COPYRIGHT file distributed with this work for additional
 * information regarding copyright ownership.
 */

// NS8

include "../../_common/rndc.key";

controls {
	inet 10.53.0.8 port @CONTROLPORT@ allow { any; } keys { rndc_key; };
};

options {
	query-source address 10.53.0.8;
	notify-source 10.53.0.8;
	transfer-source 10.53.0.8;
	port @PORT@;
	pid-file "named.pid";
	session-keyfile "session.key";
	listen-on { 10.53.0.8; };
	listen-on-v6 { none; };
	recursion no;
	notify yes;
	try-tcp-refresh no;
	notify-delay 0;
	allow-new-zones yes;
	dnssec-validation no;
};

dnssec-policy "inline" {
	keys {
		ksk key-directory lifetime unlimited algorithm @DEFAULT_ALGORITHM@;
		zsk key-directory lifetime unlimited algorithm @DEFAULT_ALGORITHM@;
	};
};

zone "example01.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example01.com.db";
};

zone "example02.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example02.com.db";
};

zone "example03.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example03.com.db";
};

zone "example04.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example04.com.db";
};

zone "example05.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example05.com.db";
};

zone "example06.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example06.com.db";
};

zone "example07.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example07.com.db";
};

zone "example08.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example08.com.db";
};

zone "example09.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example09.com.db";
};

zone "example10.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example10.com.db";
};

zone "example11.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example11.com.db";
};

zone "example12.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example12.com.db";
};

zone "example13.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example13.com.db";
};

zone "example14.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example14.com.db";
};

zone "example15.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example15.com.db";
};

zone "example16.com" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example16.com.db";
};

zone example {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "example.db";
};

zone "unsigned-serial-test" {
	type primary;
	inline-signing yes;
	dnssec-policy inline;
	file "unsigned-serial-test.db";
};
